/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/2/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_cont_tree_iterator_h__
#define __cxx_cont_tree_iterator_h__ 1


/*   std includes   */
#include <cstddef>
#include <iterator>
#include <type_traits>
#include <string>


/*   project headers   */
#include <cxx/misc/cxx-iterator.h>



namespace duct {
  namespace container {
  
    
    /*   node link for binary tree   */
    struct __tree_node_link {
      __tree_node_link *up = nullptr, *left = nullptr, *right = nullptr;
      
      /*   observers   */
      bool is_leaf() const { return (left == nullptr && right == nullptr);}
    };
    
    /*   tree node type   */
    template<class _Tp>
    struct __tree_node : __tree_node_link {
      _Tp data;
    };
    
    
    /*   return the topology of the tree given by the head node   */
    std::string topology(const __tree_node_link *);
 
    
    /*    helper class for implementing the iterators   */
    template<typename, bool> struct __tree_traverser_impl;

    
    /*   In-order iterator   */
    template<bool _is_const>
    struct __tree_traverser_impl<inorder, _is_const>
    {
      /*   link pointer type   */
      using link_ptr = std::conditional_t<_is_const, const __tree_node_link *, __tree_node_link *>;
 
      /*   increment operators   */
      static link_ptr increment(link_ptr c, link_ptr end)
      {
        if(c->right) {
          c = c->right;
          while(c->left) c = c->left;
        }
        else {
          while(c->up != end && c->up->right == c) c = c->up;
          c = c->up;
        }
        
        return c;
      }
      
      /*   decrement operators   */
      static link_ptr decrement(link_ptr c, link_ptr end)
      {
        if(c->left) {
          c = c->left;
          while(c->right) c = c->right;
        }
        else {
          while(c->up != end && c->up->left == c) c = c->up;
          c = c->up;
        }
        
        return c;
      }
      
      /*   begin operators   */
      static link_ptr  begin(link_ptr c, link_ptr end) {
        if(c) while(c->left) c = c->left;
        else c = end;
        return c;
      }
      
      /*   rbegin operators   */
      static link_ptr rbegin(link_ptr c, link_ptr end) {
        if(c) while(c->right) c = c->right;
        else c= end;
        return c;
      }
    };
    
    
    /*   Pre-order iterator    */
    template<bool _is_const>
    struct __tree_traverser_impl<preorder, _is_const>
    {
      /*   link pointer type   */
      using link_ptr = std::conditional_t<_is_const, const __tree_node_link *, __tree_node_link *>;
      
      /*   increment operator   */
      static link_ptr increment(link_ptr c, link_ptr end)
      {
        if(c->left) c = c->left;
        else if(c->right) c = c->right;
        else {
          while(c->up != end && (c->up->right == c || c->up->right == nullptr)) c = c->up;
          c = c->up != end ? c->up->right : c->up;
        }
        
        return c;
      }
      
      /*   decrement operator   */
      static link_ptr decrement(link_ptr c, link_ptr end)
      {
        if(c->up == end || c->up->left == c) c = c->up;
        else {
          c = c->up;
          if(c->left) {
            c = c->left;
            while(c->right || c->left)
              c = c->right ? c->right : c->left;
          }
        }
        
        return c;
      }

      /*   begin operators   */
      static link_ptr begin(link_ptr c, link_ptr end) {
        return (c ? c : end);
      }

      /*   rbegin operators   */
      static link_ptr rbegin(link_ptr c, link_ptr end) {
        if(c) while(c->left || c->right) c = (c->right ? c->right : c->left);
        else c = end;
        return c;
      }
   };
    
    
    /*   Post-order iterator    */
    template<bool _is_const>
    struct __tree_traverser_impl<postorder, _is_const>
    {
      /*   link pointer type   */
      using link_ptr = std::conditional_t<_is_const, const __tree_node_link *, __tree_node_link *>;
      
      /*   increment operator   */
      static link_ptr increment(link_ptr c, link_ptr end)
      {
        if(c->up == end || c->up->right == c) c = c->up;
        else {
          c = c->up;
          if(c->right) {
            c = c->right;
            while(c->left || c->right)
              c = c->left ? c->left : c->right;
          }
        }
        
        return c;
      }
      
      /*   decrement operator   */
      static link_ptr decrement(link_ptr c, link_ptr end)
      {
        if(c->right) c = c->right;
        else if(c->left) c = c->left;
        else {
          while(c->up != end && (c->up->left == c || c->up->left == nullptr)) c = c->up;
          c = c->up != end ? c->up->left : c->up;
        }
        
        return c;
      }

      /*   begin operators   */
      static link_ptr begin(link_ptr c, link_ptr end) {
        if(c) while(c->left || c->right) c = (c->left ? c->left : c->right);
        else c = end;
        return c;
      }

      /*   rbegin operators   */
      static link_ptr rbegin(link_ptr c, link_ptr end) {
        return c ? c : end;
      }
    };
    
    
    /*   leaf iterator    */
    template<bool _is_const>
    struct __tree_traverser_impl<leaf, _is_const>
    {
      /*   link pointer type   */
      using link_ptr = std::conditional_t<_is_const, const __tree_node_link *, __tree_node_link *>;
      
      /*   increment operator   */
      static link_ptr increment(link_ptr c, link_ptr end)
      {
        do {
          if(c->right) {
            c = c->right;
            while(c->left) c = c->left;
          }
          else {
            while(c->up != end && c->up->right == c) c = c->up;
            c = c->up;
          }
        } while(c != end && (c->left || c->right));
        
        return c;
      }
 
      /*   decrement operator   */
      static link_ptr decrement(link_ptr c, link_ptr end)
      {
        do {
          if(c->left) {
            c = c->left;
            while(c->right) c = c->right;
          }
          else {
            while(c->up != end && c->up->left == c) c = c->up;
            c = c->up;
          }
        } while(c != end && (c->right || c->left));
        
        return c;
      }
      
      /*   begin operators   */
      static link_ptr begin(link_ptr c, link_ptr end) {
        if(c) while(c->left) c = c->left;
        else return end;
        if(c == end || c->is_leaf()) return c;
        return increment(c, end);
      }

      /*   begin operators   */
      static link_ptr rbegin(link_ptr c, link_ptr end) {
        if(c) while(c->right) c = c->right;
        else return end;
        if(c == end || c->is_leaf()) return c;
        return decrement(c, end);
      }
    };

  

    /*   definition of the tree cursor template class  */
    template<bool _Is_Const>
    class __tree_cursor
    {
    protected:
      /*   private types   */
      using link_ptr = std::conditional_t<_Is_Const, const __tree_node_link *, __tree_node_link *>;
      
    public:
      /*   Constructors   */
      __tree_cursor() = default;
      __tree_cursor(const __tree_cursor&) = default;
      
      /*   converting the iterator to __tree_node_link *   */
      operator link_ptr() const {
        return curr;
      }
      
      /*    Comparison operators   */
      friend bool operator==(const __tree_cursor& j, const __tree_cursor& i) { return j.curr == i.curr;}
      friend bool operator!=(const __tree_cursor& j, const __tree_cursor& i) { return j.curr != i.curr;}
      
      friend bool operator==(const __tree_cursor& i, std::nullptr_t) { return i.curr == nullptr;}
      friend bool operator!=(const __tree_cursor& i, std::nullptr_t) { return i.curr != nullptr;}
      
      friend bool operator==(std::nullptr_t, const __tree_cursor& i) { return (i == nullptr);}
      friend bool operator!=(std::nullptr_t, const __tree_cursor& i) { return (i != nullptr);}
      
      /*   Binary tree specific functions   */
      __tree_cursor up() const { return __tree_cursor{root, curr->up};}
      __tree_cursor left() const { return __tree_cursor{root, curr->left};}
      __tree_cursor right() const { return __tree_cursor{root, curr->right};}
      
      /*   Information about the current node   */
      bool is_leaf() const { return curr->is_leaf();}
      bool is_root() const { return curr == root;}
      
    protected:
      /*   root and current position   */
      link_ptr root, curr;

    public:
      /*  constructors (originaly I wanted them to be protected)   */
      __tree_cursor(link_ptr r, link_ptr c) : root{r}, curr{c} {}
      __tree_cursor(const __tree_cursor<!_Is_Const>& c) : root{c.root}, curr{c.curr} {}
  
      /*   friends   */
      template<bool> friend class __tree_cursor;
    };
    
    
    /*   definition of the tree iterator template class  */
    template<class _Traversal, typename _ValueType>
    class __tree_traverser : public __tree_cursor<std::is_const<_ValueType>::value>
    {
      /*   private types   */
      static constexpr bool _Is_const = std::is_const<_ValueType>::value;
      
      using _Base = __tree_cursor<_Is_const>;
      using __node_ptr = std::conditional_t<_Is_const, const __tree_node<std::remove_const_t<_ValueType>> *, __tree_node<std::remove_const_t<_ValueType>> *>;
      
    public:
      /*   some types to be compatible with the standard   */
      using value_type = _ValueType;
      using pointer = value_type*;
      using reference = value_type&;
      
      using difference_type = std::ptrdiff_t;
      using iterator_category = std::bidirectional_iterator_tag;
      
      /*   Constructors   */
      __tree_traverser() = default;
      __tree_traverser(typename _Base::link_ptr h, typename _Base::link_ptr c) : _Base{h,c} {}
      
      /*   creating form const/non-const cursor type if it is possible   */
      template<bool _constX> __tree_traverser(const __tree_cursor<_constX>& c) : _Base{c} {}
      
      /*   Dereference operators   */
      reference operator*() const {
        return static_cast<__node_ptr>(_Base::curr)->data;
      }
      
      pointer operator->() const {
        return std::addressof(static_cast<__node_ptr>(_Base::curr)->data);
      }
      
      /*   Increment operators  */
      __tree_traverser& operator++() {
        _Base::curr = __tree_traverser_impl<_Traversal,_Is_const>::increment(_Base::curr, _Base::root);
        return *this;
      }
      
      __tree_traverser operator++(int) {
        __tree_traverser tmp = *this;
        this->operator++();
        return tmp;
      }
      
      /*   Decrement operators  */
      __tree_traverser& operator--() {
        _Base::curr = (_Base::curr == _Base::root
                       ? __tree_traverser_impl<_Traversal,_Is_const>::rbegin(_Base::root->left ? _Base::root->left : _Base::root->right, _Base::root)
                       : __tree_traverser_impl<_Traversal,_Is_const>::decrement(_Base::curr, _Base::root));
        return *this;
      }
      
      __tree_traverser operator--(int) {
        __tree_traverser tmp = *this;
        this->operator--();
        return tmp;
      }
    };
    
  }    //  namespace containers
}    //  namespace duct

#endif
