/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 24/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_stackvector_h__
#define __cxx_stackvector_h__ 1


/*   std includes   */
#include <iterator>
#include <stdexcept>
#include <algorithm>
#include <type_traits>
#include <initializer_list>

/*   project headers   */



namespace duct {
  
  
  template<class _Tp, std::size_t _BufferSize>
  requires std::is_trivially_copyable<_Tp>::value
  class stackvector
  {
    using _Storage = typename std::aligned_storage<_BufferSize*sizeof(_Tp), alignof(_Tp)>::type;
    
  public:
    /*    ususal typedefs   */
    using value_type = _Tp;
    using size_type = std::size_t;
    
    using pointer = value_type *;
    using iterator = value_type *;
    using reference = value_type&;
    
    using const_pointer = const value_type *;
    using const_iterator = const value_type *;
    using const_reference = const value_type&;
    
    using reverse_iterator = std::reverse_iterator<iterator>;
    using const_reverse_iterator = std::reverse_iterator<const_iterator>;
    
    
    /*   constructors   */
    stackvector() : _M_size{0} {}
    
    explicit stackvector(std::size_t s, const _Tp& c = _Tp())
    : _M_size(std::min(s, _BufferSize))
    {
      pointer data = reinterpret_cast<pointer>(&_M_data);
      for(std::size_t k = 0; k < _M_size; ++k)
        new(data+k) value_type(c);
    }
    
    stackvector(std::initializer_list<value_type> il)
    : _M_size(std::min(il.size(), _BufferSize))
    {
      pointer data = reinterpret_cast<pointer>(&_M_data);
      for(auto ik = il.begin(); ik < il.begin()+_M_size; ++ik)
        new(data++) value_type(*ik);
    }
    
    /*    destructor   */
    ~stackvector()
    {
      pointer data = reinterpret_cast<pointer>(&_M_data);
      for(std::size_t k = 0; k < _M_size; ++k)
        data[k].~value_type();
    }
    
    
    /*   iterators   */
    iterator begin() { return reinterpret_cast<pointer>(&_M_data);}
    iterator end() { return begin() + _M_size;}
    
    const_iterator begin() const { return reinterpret_cast<const_pointer>(&_M_data);}
    const_iterator end() const { return begin() + _M_size;}
    
    const_iterator cbegin() const { return reinterpret_cast<const_pointer>(&_M_data);}
    const_iterator cend() const { return cbegin()+_M_size;}
    
    reverse_iterator rbegin() { return reverse_iterator(end());}
    reverse_iterator rend() { return reverse_iterator(begin());}
    
    const_reverse_iterator rbegin() const { return reverse_iterator(end());}
    const_reverse_iterator rend() const { return reverse_iterator(begin());}
    
    const_reverse_iterator crbegin() const { return reverse_iterator(end());}
    const_reverse_iterator crend() const { return reverse_iterator(begin());}
    
    
    /*   resizeing the vector   */
    void resize(std::size_t new_size, const _Tp& c = _Tp())
    {
      auto __size = std::min(_BufferSize, new_size);
      
      /*   do nothing if the new size is the same   */
      if(__size == _M_size) return;
      
      if(__size < _M_size) while(_M_size > __size) pop_back();
      else while(_M_size < __size) push_back(c);
    }
    
    void push_back(const value_type& v) {
      if(_M_size < _BufferSize)
        new(reinterpret_cast<pointer>(&_M_data)+(_M_size++)) value_type(v);
    }
    
    void push_back(value_type&& v) {
      if(_M_size < _BufferSize)
        new(reinterpret_cast<pointer>(&_M_data)+(_M_size++)) value_type(std::move(v));
    }
    
    template<typename... _Args>
    void emplace_back(_Args&&... args) {
      if(_M_size < _BufferSize)
        new(reinterpret_cast<pointer>(&_M_data)+(_M_size++)) value_type(std::forward<_Args>(args)...);
    }
    
    void pop_back() {
      reinterpret_cast<pointer>(&_M_data)[--_M_size].~value_type();
    }
    
    /*   remove all the elements   */
    void clear() { while(_M_size) pop_back();}
    
    
    /*   element access    */
    reference operator[](std::size_t k) { return reinterpret_cast<pointer>(&_M_data)[k];}
    constexpr const_reference operator[](std::size_t k) const { return reinterpret_cast<const_pointer>(&_M_data)[k];}
    
    reference at(std::size_t k) {
      if(k < _M_size) return reinterpret_cast<pointer>(&_M_data)[k];
      else throw std::out_of_range{"duct::stackvector<...> : out of range"};
    }
    
    const_reference at(std::size_t k) const {
      if(k < _M_size) return reinterpret_cast<const_pointer>(&_M_data)[k];
      else throw std::out_of_range{"duct::stackvector<...> : out of range"};
    }
    
    pointer data() { return reinterpret_cast<pointer>(&_M_data);}
    const_pointer data() const { return reinterpret_cast<const_pointer>(&_M_data);}
    
    /*   structural informations   */
    constexpr std::size_t size() const { return _M_size;}
    constexpr std::size_t capacity() const { return _BufferSize;}
    constexpr bool empty() const { return !_M_size;}
    
  private:
    /*   actual size of the vector   */
    std::size_t _M_size;
    
    /*   the underlying buffer   */
    _Storage _M_data;
  };
  
  
}

#endif
