/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1



/*   std includes   */
#include <cmath>
#include <utility>
#include <fstream>
#include <string>
#include <stack>


/*   project headers   */
#include <bits/mc-shower-history.h>
#include <bits/mc-hard-params.h>
#include <bits/mc-splitting-prob.h>
#include <bits/qcd-color-overlap.h>
#include <bits/mc-engine-opts.h>
#include <cxx/cloned-ptr.h>



namespace duct {
  
  namespace __details {
    
    
    /*  This function selects the color operator. This can be different in different
     *  shower implementations. This function select the color operator in dipole shower.
     */
    double selecting_color_operator_su3(random_engine_type&, splitting_vars *, const particle_state *, const model_ref&);
    double selecting_color_operator_u3(random_engine_type&, splitting_vars *, const particle_state *, const model_ref&);
    double selecting_color_operator_su3_ggg(random_engine_type&, splitting_vars *, splitting_vars *, const particle_state *);
    
    
    template<class _Proc> inline
    double selecting_color_operator_su3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_su3(rne, std::addressof(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
    
    template<class _Proc> inline
    double selecting_color_operator_u3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_u3(rne, std::addressof(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
 
    
    template<class _Proc> inline
    double selecting_color_operator_su3_ggg(random_engine_type& rne, shower_stage<_Proc>& stage, splitting_vars& branch) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_su3_ggg(rne, std::addressof(stage.vars), &branch, qcd.begin()-1);
    }
 
  }    //   namespace __details

  
  /*   Requirements for a shower probability that can make trial emission   */
  template<typename _Prob, ProcessTag _Proc = typename _Prob::process_type>
  concept bool ShowerProbability =
  requires(_Prob& s, const shower_engine_opts& opts, shower_stage<_Proc>& stage, random_engine_type& rne, double  tmH, double tmL)
  {
    /*   type members    */
    typename _Prob::process_type;
    typename _Prob::ordering_type;

    /*   type constraints   */
    requires std::is_same<_Proc, typename _Prob::process_type>::value;
    
    /*  copy constructable   */
    {_Prob{const_cast<const _Prob&>(s)}};
    
    /*   can make a trial emission (at least the function is there for this)   */
    {s.make_trial(opts, stage, rne, tmH, tmL)}  -> bool;
    
    /*   there is function for calculating the threshold weight    */
    {s.threshold(opts, stage)} -> __details::thrs_weight;
  };
  
  
  /*  This is the shower evolution code. It uses a given shower probability class to make
   *  trial emission. An instance of the shower probability class is stored in this
   *  object. The type of the probability class is erased.
   */
  template<ProcessTag _Proc>
  class shower_engine
  {
    /*   private types   */
    using _History = shower_history<_Proc>;
    using _Iterator = typename _History::iterator;
    
    
  public:
    /*    default constructor    */
    shower_engine() = default;
    
    /*    create the shower engine from a given shower probability   */
    template<class _ShowerProb>
    requires ShowerProbability<_ShowerProb, _Proc>
    explicit shower_engine(_ShowerProb&& prob)
    : _M_trial{new obj<_ShowerProb>{std::forward<_ShowerProb>(prob)}} {}
    
    
    /*   shower evolution (at the moment it is just the LC+ approximation)   */
    void evolve(double tmax, shower_history<_Proc>& hist) {
      _M_evolve_lcplus(hist, tmax);
    }

    /*   same like evolve but it returns the history   */
    shower_history<_Proc>& operator()(double tmax, shower_history<_Proc>& hist) {
      this->evolve(tmax, hist);
      return hist;
    }
    
    
    /*    calculating the threshold weight    */
    void apply_threshold_weights(shower_history<_Proc>& hist) const
    {
      if(is_hadron_collision<_Proc>) {
        if(_M_opts.threshold.enabled)
        {
          for(auto& stage : hist.hard())
            stage.weights.apply_threshold_correction(_M_trial -> threshold(_M_opts, stage));
        }
      }
    }
    
    
    /*   Setting/getting engine options     */
    void options(const shower_engine_opts& opts) { _M_opts = opts;}
    const shower_engine_opts& options() const { return _M_opts;}
    
    /*   set the random number seed   */
    void seed(random_engine_type::result_type s) { _M_random.seed(s);}
    
  private:
    /*   random engine   */
    random_engine_type _M_random;
    
    /*   engine parameters    */
    shower_engine_opts _M_opts;
    
    /*   calculates color overlaps in the evolution    */
    color_overlap_su3 _M_overlap_su3;
    color_overlap_u3 _M_overlap_u3;
    
    /*   some workspace for keep tracking of the unshowered nodes   */
    std::stack<_Iterator> _M_not_showered;
    
    
  private:
    /*    make one shower step   */
    _Iterator _M_make_one_step(shower_history<_Proc>&, _Iterator, double);
    
    /*   LC+ approximation of the shower evolution     */
    void _M_evolve_lcplus(shower_history<_Proc>&, double);
    
    
    /*   abstract class for type erasing   */
    struct base {
      virtual bool make_trial(const shower_engine_opts&, shower_stage<_Proc>&, random_engine_type&, double, double) = 0;
      virtual __details::thrs_weight threshold(const shower_engine_opts&, const shower_stage<_Proc>&) = 0;
    };
    
    
    /*   this stores an instanse of the actual splitting probability   */
    template<class _ShowerProb>
    struct obj : base
    {
      explicit obj(_ShowerProb&& s) : prob{std::forward<_ShowerProb>(s)} {}
      
      bool make_trial(const shower_engine_opts& opts, shower_stage<_Proc>& s, random_engine_type& r, double tmU, double tmL) {
        return prob.make_trial(opts, s, r, tmU, tmL);
      }
      
      __details::thrs_weight threshold(const shower_engine_opts& opts, const shower_stage<_Proc>& stage) {
        return prob.threshold(opts, stage);
      }
      
      _ShowerProb prob;
    };
    
    
    /*   we put an instance of obj into an cloned_ptr   */
    cloned_ptr<base> _M_trial;
    
    
    /*   change color group (SU(Nc) ==> U(Nc))    */
    void _M_change_color_group(_Iterator next)
    {
      if(next->is_color_u3 == false && next->color_suppression_index >= _M_opts.max_color_suppression)
      {
        next->weights.unitary.main *= _M_overlap_su3(next->state);
        next->weights.unitary.main /= _M_overlap_u3(next->state);
        next->is_color_u3 = true;
      }
    }
  
    
    /*   retirns the color supression index of the next state   */
    static unsigned int _S_next_color_suppression_index_lcplus(typename shower_history<_Proc>::const_iterator current)
    {
      const auto& vars = current->vars;

      if(current->state[vars.emitter].health == 'h')
      {
        /*    healthy gluon, crossed configuration    */
        if(std::abs(vars.color.sigma) == 1 && vars.color.sigma == -vars.colorP.sigma)
          return current->color_suppression_index + 2u;
        
        /*    gluon splits to quark-antiquark    */
        if(vars.color.sigma == -2 || vars.colorP.sigma == -2)
          return current->color_suppression_index + 2u;
        
        /*    gluon splits to antiquark-quark    */
        if(vars.color.sigma == -3 || vars.colorP.sigma == -3)
          return current->color_suppression_index + 2u;
      }
      
      return current->color_suppression_index;
    }
  };
  
 
  
  
  
  template<ProcessTag _Proc>
  auto shower_engine<_Proc>::_M_make_one_step(shower_history<_Proc>& hist, _Iterator current, double tmax) -> _Iterator
  {
    /*     make trial shower emissions    */
    bool is_possible = _M_trial -> make_trial(_M_opts, *current, _M_random, tmax, current->current_time);
    
    /*   if there is no splitting we return with the current state    */
    if(!is_possible) return current;
    

    /*
     *    COLOR OPERATOR : LC+ contributions
     */
    const model_ref &mdl = current->state.get_model();

    auto& vars = current->vars;
    int cl = mdl.color(vars.fl), cm = mdl.color(vars.fm);
 
    
    /*   emitter is a frail gluon and it is a g -> g + g splitting   */
    if(_M_opts.allow_color_branching
       && current->is_color_u3 == false
       && current->state[vars.emitter].health == 'f'
       && cl == 8 && cm == 8)
    {
      /*   make the mother state of the new branch    */
      _Iterator branch = hist.append_branch(current, *current);
      
      /*   select the "composite" color operator   */
      double weight = __details::selecting_color_operator_su3_ggg(_M_random, *current, branch->vars);

      
      /*    appending an uninitialized node to the history     */
      _Iterator next = hist.append(current);
      
      next->weights = current->weights;
      next->weights.unitary.main *= weight;
      
      next->is_color_u3 = current->is_color_u3;
      next->current_time = current->next_time;
      next->next_time = tmax;
      next->vetos = current->vetos;
      next->hard_color_suppression_index = current->hard_color_suppression_index;
      
      /*     assigning the new momenta and flavors    */
      next->qnull = shower_mapping_no_color(next->state, current->state, current->vars)(current->qnull);
      
      /*   the scale of the shower state   */
      next->scale = next->qnull.mag2()*std::exp(-next->current_time);

      /*   appending the new state on the other branch   */
      _Iterator nextB = hist.append(branch, *next);
      nextB->weights.unitary.main *= -1.0;

      
      /*   mapping the color states    */
      shower_color_mapping(next->state, current->state, current->vars);
      shower_color_mapping(nextB->state, branch->state, branch->vars);

      /*   color supression index    */
      next->color_suppression_index  = _S_next_color_suppression_index_lcplus(current);
      nextB->color_suppression_index = _S_next_color_suppression_index_lcplus(branch);
      
      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next);
      _M_change_color_group(nextB);
      
      /*   keep showering with the branch of next and do the evolution for the branch of "branch" later   */
      if(std::abs(nextB->weights.unitary.main) > 0.0)
        _M_not_showered.push(nextB);
      
 
      return next;
    }
    
    /*   emitter is a frail gluon and it is a g -> q + q~ splitting   */
    else if(_M_opts.allow_color_branching
            && current->is_color_u3 == false
            && current->state[vars.emitter].health == 'f'
            && std::abs(cl) == 3 && cm == -cl)
    {
      /*   make the mother state of the new branch    */
      _Iterator branch[3];
      
      branch[0] = hist.append_branch(current, *current);
      branch[1] = hist.append_branch(current, *current);
      branch[2] = hist.append_branch(current, *current);
      
      
      /*    fixing color vertices in the mother states   */
      int sigma = (cl == 3 ? 2 : 3), emit = vars.emitter;

      vars.color  = {sigma, emit};
      vars.colorP = {sigma, emit};
 
      branch[0]->vars.color  = { sigma, emit};
      branch[0]->vars.colorP = {-sigma, emit};
      
      branch[1]->vars.color  = {-sigma, emit};
      branch[1]->vars.colorP = { sigma, emit};
    
      branch[2]->vars.color  = {-sigma, emit};
      branch[2]->vars.colorP = {-sigma, emit};


      /*    appending an uninitialized node to the history     */
      _Iterator next[4];
      
      next[0] = hist.append(current);
      
      next[0]->weights = current->weights;
      next[0]->is_color_u3 = current->is_color_u3;
      next[0]->current_time = current->next_time;
      next[0]->next_time = tmax;
      next[0]->vetos = current->vetos;
      next[0]->hard_color_suppression_index = current->hard_color_suppression_index;
      
      
      /*   assigning the new momenta and flavors    */
      next[0]->qnull = shower_mapping_no_color(next[0]->state, current->state, current->vars)(current->qnull);
      
      /*   the scale of the shower state   */
      next[0]->scale = next[0]->qnull.mag2()*std::exp(-next[0]->current_time);

      
      /*   appending the new state on the other branch   */
      next[1] = hist.append(branch[0], *(next[0]));
      next[2] = hist.append(branch[1], *(next[0]));
      next[3] = hist.append(branch[2], *(next[0]));
      
      
      /*   fixing the weights   */
      next[0]->weights.unitary.main *=  0.5*qcd::Nc/qcd::Cf;
      next[1]->weights.unitary.main *= -0.5/qcd::Cf;
      next[2]->weights.unitary.main *= -0.5/qcd::Cf;
      next[3]->weights.unitary.main *=  0.5/(qcd::Nc*qcd::Cf);
      
      
      /*   mapping the color states    */
      shower_color_mapping(next[0]->state, current->state, current->vars);
      shower_color_mapping(next[1]->state, branch[0]->state, branch[0]->vars);
      shower_color_mapping(next[2]->state, branch[1]->state, branch[1]->vars);
      shower_color_mapping(next[3]->state, branch[2]->state, branch[2]->vars);
      
      /*   color supression index    */
      next[0]->color_suppression_index = _S_next_color_suppression_index_lcplus(current);
      next[1]->color_suppression_index = _S_next_color_suppression_index_lcplus(branch[0]);
      next[2]->color_suppression_index = _S_next_color_suppression_index_lcplus(branch[1]);
      next[3]->color_suppression_index = _S_next_color_suppression_index_lcplus(branch[2]);
      
      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next[0]);
      _M_change_color_group(next[1]);
      _M_change_color_group(next[2]);
      _M_change_color_group(next[3]);
      
      
      /*   keep showering with the branch of next and do the evolution for the branch of "branch" later   */
      for(int k : {1,2,3})
        if(std::abs(next[k]->weights.unitary.main) > 0.0)
          _M_not_showered.push(next[k]);
          
          
      return next[0];
    }
    
    /*   no brancing in the history, we have the simple Markovian chain   */
    else
    {
      /*    selecting color operator    */
      double weight = 0.0;
    
      if(current->is_color_u3) weight = __details::selecting_color_operator_u3(_M_random, *current);
      else weight = __details::selecting_color_operator_su3(_M_random, *current);
    
      /*    appending an uninitialized node to the history     */
      _Iterator next = hist.append(current);
      
      next->weights = current->weights;
      next->weights.unitary.main *= weight;
      
      next->is_color_u3 = current->is_color_u3;
      next->current_time = current->next_time;
      next->next_time = tmax;
      next->vetos = current->vetos;
      next->hard_color_suppression_index = current->hard_color_suppression_index;
      
      /*     assigning the new momenta and flavors    */
      next->qnull = shower_mapping(next->state, current->state, current->vars)(current->qnull);
      
      /*   the scale of the shower state   */
      next->scale = next->qnull.mag2()*std::exp(-next->current_time);

      /*   color supression index    */
      next->color_suppression_index = _S_next_color_suppression_index_lcplus(current);

      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next);
      
      return next;
    }
  }
  
  
 
  template<ProcessTag _Proc>
  void shower_engine<_Proc>::_M_evolve_lcplus(shower_history<_Proc>& hist, double tmF)
  {
    /*    save all the leaf node of the tree   */
    for(auto ip = hist.begin(); ip != hist.end(); ++ip)
      if(tmF > ip->next_time && std::abs(ip->weights.unitary.main) > 0.0)
      {
        /*    change color group of the starting stage, SU(Nc) ==> U(Nc), if it is needed   */
        _M_change_color_group(ip);
        
        /*   if the weight is not zero we enroll this stage for shower evolution    */
        if(std::abs(ip->weights.unitary.main) > 0.0)
          _M_not_showered.push(ip);
      }
    
    
    /*   shower every stages   */
    while(_M_not_showered.empty() == false)
    {
      /*   pop the top of the stack   */
      _Iterator next = _M_not_showered.top(); _M_not_showered.pop();
      
      
      /*    start the evolution   */
      _Iterator current;
      
      do {
        current = next;
        next = _M_make_one_step(hist, current, tmF);
      } while(current != next && std::abs(next->weights.unitary.main) > 0.0);
    }
  }
    
  

}   /*   namespace duct   */

#endif
