/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_user_h__
#define __mc_user_h__ 1


/*   project headers     */
#include <bits/mc-engine.h>
#include <bits/qcd-color-overlap.h>


namespace duct {
  
  /*   Abstract class to define the most general user interface.
   *   This is just the base of the user class, but it is independent 
   *   of the process type.
   */
  class user_base
  {
  public:
    /*   Virtual destructor   */
    virtual ~user_base() = default;
    
    /*   User defined finish function. It is called once
     *   at the end of the calculation.
     */
    virtual void finish() {}
    
    /*   Using this function the deductor kernel can pass some information
     *   such as PID and name of the process. Those are helpful to compose 
     *   output file names. The default does nothing.
     */
    virtual void process_info(unsigned int, const char *) {}

    /*   Make sure that the event counter is set to zero
     *   and the maximum number of events is set correctly.
     */
    void initialize() {
      this->initfunc();
      this->_M_number_of_events = 0;
    }
    
    /*   Increment the event number and do some user defined operations at the
     *   end of the event
     */
    void end_of_event() {
      ++_M_number_of_events;
      this -> operations_at_the_end_of_event();
    }
    
    /*   Returns the number of the analysed events    */
    unsigned long long int number_of_events() const {
      return _M_number_of_events;
    }
    
    /*   returns true if this calculation is finished (no more events are needed)   */
    bool is_done() const {
      return max_number_of_events == 0 ? false : _M_number_of_events >= max_number_of_events;
    }
    
  private:
    /*   User defined init function. It is called once
     *   at the begining of the calculation.
     */
    virtual void initfunc() = 0;
    
    /*   User defined operations at the end of the event. It is virtual but not abstract
     *   function. User might don't have to do anything at the end of the event.
     */
    virtual void operations_at_the_end_of_event() {}
    
  protected:
    /*   maximum number of events to generate  */
    unsigned long long int max_number_of_events = 0;
    
  private:
    /*   number of the analysed events   */
    unsigned long long int _M_number_of_events = 0;
    
  };
  
  
  /*   Abstract class to define the most general user interface.
   *   Every analyser should be inhereted from this class.
   *   This class adds the userfunc(...) function and defines the
   *   full user<Tag> abstract class.
   */
  template<ProcessTag _Tag>
  class user : public virtual user_base
  {
  public:
    /*   useful aliases in the analysers   */
    using weight_type = typename weight_set<_Tag>::user_weight_type;
    
    /*   Virtual destructor   */
    virtual ~user() = default;
    
    /*   The analyser routine. It analyses the events and fill the histograms.   */
    virtual void userfunc(const shower_history<_Tag>&) = 0;

    /*   This is the overlap calculator function.    */
    double color_weight(const shower_stage<_Tag>& s) const {
      if(s.is_color_u3) return _M_overlap_u3(s.state);
      else return _M_overlap_su3(s.state);
    }

    /*   calculates the weight with NLO PDF at hard level:
     *      0 => full weight with all the threshold weight
     *      1 => no Delta contribution
     *      2 => no Delta and Preg contribution
     *      3 => no threshold at all
     */
    typename weight_set<_Tag>::user_weight_type
    weight(const shower_stage<_Tag>& stage, double fac = 1.0) const {
      return stage.weights.user_weight_nlo_pdf(color_weight(stage)*fac);
    }
    
    /*   calculates the weight with LO PDF at hard level:
     *      0 => full weight with all the threshold weight
     *      1 => no Delta contribution
     *      2 => no Delta and Preg contribution
     *      3 => no threshold at all
     */
    typename weight_set<_Tag>::user_weight_type
    weight_lo_pdf(const shower_stage<_Tag>& stage, double fac = 1.0) const {
      return stage.weights.user_weight_lo_pdf(color_weight(stage)*fac);
    }
    
  private:
    /*   every analyser has its own resources to calculate color overlaps   */
    mutable color_overlap_su3 _M_overlap_su3;
    mutable color_overlap_u3 _M_overlap_u3;
  };


}

#endif
