/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 7/4/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <climits>
#include <fstream>

/*   project headers   */
#include "bits/duct-model.h"


namespace duct {
  
  
  
  static const particle_info __particle_table_three_light_quarks[] = {
    /*   standard quark flavors   */
    {-6,      "-t",      175.0,        -3, 1},
    {-5,      "-b",        4.75,       -3, 1},
    {-4,      "-c",        1.4,        -3, 1},
    {-3,      "-s",        0.0,        -3, 1},
    {-2,      "-u",        0.0,        -3, 1},
    {-1,      "-d",        0.0,        -3, 1},
    { 0,      "gluon",     0.0,         8, 2},
    { 1,      "d",         0.0,         3, 1},
    { 2,      "u",         0.0,         3, 1},
    { 3,      "s",         0.0,         3, 1},
    { 4,      "c",         1.4,         3, 1},
    { 5,      "b",         4.75,        3, 1},
    { 6,      "t",       175.0,         3, 1},
 
    /*   leptons   */
    { 11,     "e-",        0.0,         0, 1},
    {-11,     "e+",        0.0,         0, 1},
    { 12,     "nu_e",      0.0,         0, 1},
    {-12,     "-nu_e",     0.0,         0, 1},

    { 13,     "mu-",       0.0,         0, 1},
    {-13,     "mu+",       0.0,         0, 1},
    { 14,     "nu_mu",     0.0,         0, 1},
    {-14,     "-nu_mu",    0.0,         0, 1},

    { 15,     "tau-",      1.77682,     0, 1},
    {-15,     "tau+",      1.77682,     0, 1},
    { 16,     "nu_tau",    0.0,         0, 1},
    {-16,     "-nu_tau",   0.0,         0, 1},
    
    /*   gauge bosons   */
    { 21,     "gluon",     0.0,         8, 2},
    { 22,     "photon",    0.0,         0, 2},
    { 23,     "Z0",       91.182,       0, 2},
    { 24,     "W+",       80.4,         0, 2},
    {-24,     "W-",       80.4,         0, 2},
    { 25,     "h0",      125.0,         0, 0},
    
    /*   MC internals   */
    {-83,     "generic -b",    4.75,   -3, 1},
    {-82,     "generic -c",    1.4,    -3, 1},
    {-81,     "generic -lq",   0.0,    -3, 1},

    { 81,     "generic lq",    0.0,     3, 1},
    { 82,     "generic c",     1.4,     3, 1},
    { 83,     "generic b",     4.75,    3, 1},
    
    /*   end of list   */
    {INT_MAX, "",          0.0,         0, 0}
  };
 
  static const hadron_info __hadron_table_some_light_hadron[] = {
    { 2212, "proton"},
    {-2212, "anti-proton"},
    { 2112, "neutron"},
    {INT_MAX, ""}
  };
  
  
  model::model()
  : model{__particle_table_three_light_quarks, __hadron_table_some_light_hadron} {}
  
  
  model::model(const particle_info *pi, const hadron_info *hi)
  {
    if(pi != nullptr)
      while(pi->key != INT_MAX) {
        particle_table[pi->key] = *pi;
        ++pi;
      }
    
    if(hi != nullptr)
      while(hi->key != INT_MAX) {
        hadron_table[hi->key] = *hi;
        ++hi;
      }
  }
  
  
  void model::qcd_msbar_pdf(int key, const qcd::input& qin)
  {
    if(qin.pdf.empty() == false)
    {
      for(auto& h : qin.pdf)
        if(key == h.first) {
          /*    MSbar PDFs (LO & NLO)    */
          hadron_table[key].pdf[0].evolve<qcd::order::lo>(key, qin);
          hadron_table[-key].pdf[0] = hadron_table[key].pdf[0];
          hadron_table[-key].pdf[0].anti();
          
          hadron_table[key].pdf[1].evolve<qcd::order::nlo>(key, qin);
          hadron_table[-key].pdf[1] = hadron_table[key].pdf[1];
          hadron_table[-key].pdf[1].anti();
  
          /*   just for safety if we provide a hadron that is not in the list   */
          hadron_table[key].key = key;
          hadron_table[-key].key = -key;
        }
    }
  }
  
  void model::qcd_msbar_pdf(const qcd::input& qin)
  {
    if(qin.pdf.empty() == false)
    {
      for(auto& h : qin.pdf)
      {
        int key = h.first;
        
        /*    MSbar PDFs (LO & NLO)    */
        hadron_table[key].pdf[0].evolve<qcd::order::lo>(key, qin);
        hadron_table[-key].pdf[0] = hadron_table[key].pdf[0];
        hadron_table[-key].pdf[0].anti();
        
        hadron_table[key].pdf[1].evolve<qcd::order::lo>(key, qin);
        hadron_table[-key].pdf[1] = hadron_table[key].pdf[1];
        hadron_table[-key].pdf[1].anti();
        
        /*   just for safety if we provide a hadron that is not in the list   */
        hadron_table[key].key = key;
        hadron_table[-key].key = -key;
      }
    }
  }
  
  
  
  
  void model::write_core(std::ostream& os) const
  {
    /*   write out the sizes of the tables   */
    duct::write(os, static_cast<unsigned>(particle_table.size()), static_cast<unsigned>(hadron_table.size()));
    
    /*   writting out the particle table   */
    for(const auto& p : particle_table) {
      const auto& pi = p.second;
      duct::write(os, pi.key, pi.mass, pi.mass2, pi.color, pi.spin, pi.name);
    }
    
    /*   writting out the alpha_qcd   */
    alpha_qcd.write(os);
    
    /*   writting out the hadron table   */
    for(const auto& p : hadron_table)
      duct::write(os, p.second.key, p.second.name);
  }
  
  void model::read_core(std::istream& is)
  {
    /*   cleaning up   */
    this->particle_table.clear();
    this->hadron_table.clear();
    
    /*   reading the size of tables   */
    unsigned np = 0u, nh = 0u;
    duct::read(is, np, nh);
    
    /*   reading back the particle table   */
    particle_info pi;
    
    for(unsigned k = 0; k < np; k++) {
      duct::read(is, pi.key, pi.mass, pi.mass2, pi.color, pi.spin, pi.name);
      particle_table[pi.key] = pi;
    }
    
    /*   reading back the alpha_qcd   */
    alpha_qcd.read(is);
    
    /*   reading back the hadron table   */
    hadron_info hi;
    
    for(unsigned k = 0; k < nh; k++) {
      duct::read(is, hi.key, hi.name);
      hadron_table[hi.key] = hi;
    }
  }


  void model::read_pdf(int key, std::istream& is)
  {
    std::array<qcd::dglap, 2> pdf;
 
    duct::read(is, pdf);
    hadron_table[-key].pdf = hadron_table[key].pdf = pdf;

    hadron_table[-key].pdf[0].anti();
    hadron_table[-key].pdf[1].anti();
    
    /*   just for safety if we provide a hadron that is not in the list   */
    hadron_table[key].key = key;
    hadron_table[-key].key = -key;
  }
  
  
  model::model(fs::path dir, const qcd::input& __in) : model{}
  {
    dir /= __in.name;
    
    /*   create the directory if it doesn't exists    */
    if(!fs::is_directory(dir))
      fs::create_directories(dir);
    
    
    /*   CORE part     */
    fs::path fn{dir/"core.mdl"};
    
    /*   if we have core.mdl file then we read it   */
    if(fs::exists(fn)) {
      std::fstream file{fn.c_str(), std::ios::in};
      read_core(file);
    }
    
    /*   otherwise save the default     */
    else {
      std::fstream file{fn.c_str(), std::ios::out};
      qcd_core(__in);
      write_core(file);
    }
    
    
    /*   PDFs
     *   we read or create if there is any in the qcd::input
     */
    if(__in.pdf.empty() == false)
    {
      for(auto& h : __in.pdf)
      {
        /*   MSbar PDFs     */
        int key = h.first;
        std::string pdg{"msbar-"};
        
        pdg += std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_pdf(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_msbar_pdf(key, __in);
          duct::write(file, hadron_table.at(key).pdf);
        }
        
        
        /*   Shower PDFs: kT ordering    */
        pdg = ordering::kT::name;
        pdg += "-" + std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_aux<ordering::kT>(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_shower_aux<ordering::kT>(key, __in);
          hadron_table.at(key).aux[static_cast<unsigned>(ordering::kT::alpha)].write(file);
        }
        
        
        /*   Shower PDFs: lambda ordering    */
        pdg = ordering::lambda::name;
        pdg += "-" + std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_aux<ordering::lambda>(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_shower_aux<ordering::lambda>(key, __in);
          hadron_table.at(key).aux[static_cast<unsigned>(ordering::lambda::alpha)].write(file);
        }
        
        
        /*   Shower PDFs: angular ordering    */
        pdg = ordering::angular::name;
        pdg += "-" + std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_aux<ordering::angular>(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_shower_aux<ordering::angular>(key, __in);
          hadron_table.at(key).aux[static_cast<unsigned>(ordering::angular::alpha)].write(file);
        }
        
      }
    }
  }
  

  
  /*    We have to provide a default model and some functions to get access to it.
   */
  const model _G_default_model_singleton;
  const model *_G_default_model = &_G_default_model_singleton;
  
  
  const model * get_default_model() {
    return _G_default_model;
  }
  
  void set_default_model(const model *__new_def) {
    _G_default_model = __new_def ? __new_def : &_G_default_model_singleton;
  }

  
  
  /*   Ewlwctroweak coupling    */
  double model::alpha_qed(double Q2) const
  {
    double rpigg, alem = 1.0/137.04;
    double alem3pi = alem/(3.0*M_PI);
    
    if(Q2 <= 2e-6) rpigg = 0.0;
    else if(Q2 < 0.09) rpigg = alem3pi*(13.4916 + std::log(Q2)) + 0.00835*std::log(1.0+Q2);
    else if(Q2 < 9.0) rpigg = alem3pi*(16.3200 + 2.0*std::log(Q2)) + 0.00238*std::log(1.0+3.927*Q2);
    else if (Q2 < 10000.0) rpigg = alem3pi*(13.4955 + 3.0*std::log(Q2)) + 0.00165 + 0.00299*std::log(1.0 + Q2);
    else rpigg = alem3pi*(13.4955 + 3.0*std::log(Q2)) + 0.00221 + 0.00293*std::log(1.0 + Q2);
    
    return alem/(1.0-rpigg);
  }


}




