/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 12/03/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers   */
#include "bits/mc-splitting-kernel.h"



namespace duct {
  
  template<>
  double splitting_kernel<fin, ordering::angular>::
  hit_or_miss_ratio(const hard_params<fin> *pars, double tau, double chi)
  {
    /*    hardness variable   */
    double v = pars->mdl.alpha_qcd().q2(tau)/pars->pQ;
    
    /*   binding to the current hardness variable   */
    _M_approx.bind(v);
    
    
    /*    initialize the probabilities    */
    _M_zeta.clear();
    
    /*   We have to consider all the possible channels both in the low and high z regions  */
    double weight = 0.0;
    std::pair<double,double> z;
    
    /*   Low z region   */
    {
      /*   soft contribution    */
      z = _M_approx.mappingL_chi2z(splitting_type::soft, chi);
    
      if(_Split::veto(pars, v, z.first) && z.second > 0.0)
      {
        weight += z.second*_M_sfunc[0](splitting_type::soft, pars, v, z.first);
        
        for(const auto& s : _M_sfunc[0].soft_states())
          if(s.weight() > 0.0) _M_zeta.push_back(_Channel{z.second*s.weight(), pars->fl, 0, s.spectator(), v, z.first, &s});
      }
      
      /*   collinear contribution    */
      z = _M_approx.mappingL_chi2z(splitting_type::collinear, chi);
     
      if(_Split::veto(pars, v, z.first) && z.second > 0.0)
      {
        weight += z.second*_M_sfunc[0](splitting_type::collinear, pars, v, z.first);
        
        const auto& cw = _M_sfunc[0].collinear_weights();
        
        for(int k = 0; k < (int) cw.size(); ++k)
          if(cw[k] > 0.0) _M_zeta.push_back(_Channel{z.second*cw[k], pars->fl+k, -k, pars->emitter, v, z.first, nullptr});
      }
    }
    
    
    /*   High z region   */
    {
      /*   soft contribution    */
      z = _M_approx.mappingH_chi2z(splitting_type::soft, chi);
 
      if(_Split::veto(pars, v, z.first) && z.second > 0.0)
      {
        weight += z.second*_M_sfunc[1](splitting_type::soft, pars, v, z.first);
        
        for(const auto& s : _M_sfunc[1].soft_states())
          if(s.weight() > 0.0) _M_zeta.push_back(_Channel{z.second*s.weight(), pars->fl, 0, s.spectator(), v, z.first, &s});
      }
      
      /*   collinear contribution    */
      z = _M_approx.mappingH_chi2z(splitting_type::collinear, chi);
     
      if(_Split::veto(pars, v, z.first) && z.second > 0.0)
      {
        weight += z.second*_M_sfunc[1](splitting_type::collinear, pars, v, z.first);
        
        const auto& cw = _M_sfunc[1].collinear_weights();
        
        for(int k = 0; k < (int) cw.size(); ++k)
          if(cw[k] > 0.0) _M_zeta.push_back(_Channel{z.second*cw[k], pars->fl+k, -k, pars->emitter, v, z.first, nullptr});
      }
    }
    
    
    /*   the trial emission was vetoed    */
    if(0 == _M_zeta.size()) return 0.0;
    
    /*   normalizing the probabilities    */
    for(auto& c : _M_zeta) c.prob /= weight;
    
    
    /*   returning the hit-or-miss ratio   */
    return weight/_M_approx.integral();
  }

  
}
