/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/23/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __proc_hccjets_h__
#define __proc_hccjets_h__ 1

//   std includes

//   project headers
#include <bits/duct-model.h>
#include <bits/mc-shower-history.h>


namespace duct {
  
  class hhcjets
  {
  public:
    /*     constructor    */
    hhcjets(const model_ref&, double, std::initializer_list<double>, double = 1.0, double = 1.0, int = 2212, int = 2212);
    
 
    /*  this operator generates a new hard event   */
    void generate(random_engine_type&, shower_history<hhc>&);
  
    
  private:
    /*   pointer to the model  */
    model_ref _M_mdl;
    
    /*   ponter to the pdfs   */
    int _M_hadronA, _M_hadronB;
    const qcd::dglap *_M_pdf1, *_M_pdf2;
    
    /*   genreate the event (most of the work is done here)    */
    double _M_generate_event(random_engine_type&, shower_history<hhc>::iterator);
    
    /*    select the active subprocess    */
    double _M_select_subprocess(double, double, double);
    
    
    /*   p2T ranges  */
    std::vector<double> _M_p2T;
    mutable double _M_p2Tmin, _M_p2Tmax;
    
    /*     structure for represnting the subprocesses    */
    struct subprocess {
      int fl[4];
      double (hhcjets::*weight_function)(double, double, const int *) const;
      double (*select_color)(shower_history<hhc>::iterator, double);
    };
    
    /*     table of the possible subprocesses     */
    static const subprocess _M_proc_tbl[250];
    
    /*     cached variables to communicate between the private member functions     */
    double _M_s, _M_scalefac_pert, _M_scalefac_shw, _M_nu[7], _M_nu1[7];
    mutable double _M_fa[11], _M_fb[11];
    mutable const subprocess *_M_active;

    /*      tables of color supression indices    */
    static const unsigned int _M_color_suppr_idx_4g[6][6];
    static const unsigned int _M_color_suppr_idx_2q2g[2][2];
    static const unsigned int _M_color_suppr_idx_4q[2][2];
    static const unsigned int _M_color_suppr_idx_4qC[2][2];
    
    
    /*  Private functions help to compute the weight 
     *  of the subprocesses and select color states.
     */
    
    /*     transverse momentum     */
    double _M_transverse_momentum(double tau, double x, const int *fl) const {
      return _M_s*tau*x*(1.0-x) - _M_mdl.mass2(fl[3])*x - _M_mdl.mass2(fl[2])*(1.0-x);
    }

    /*     generating the x variable     */
    void _M_generate_x(random_engine_type&, double&, double&, double, double);
    
    /*      weight functions of the possible subprocesses     */
    double _M_process_gggg(double, double, const int *) const;
    double _M_process_ggqQ(double, double, const int *) const;
    double _M_process_gqgq(double, double, const int *) const;
    double _M_process_qggq(double, double, const int *) const;
    double _M_process_qQgg(double, double, const int *) const;
    double _M_process_qQrR(double, double, const int *) const;
    double _M_process_qrqr(double, double, const int *) const;
    double _M_processA_qQqQ(double, double, const int *) const;
    double _M_processB_qQqQ(double, double, const int *) const;
    double _M_processC_qQqQ(double, double, const int *) const;
    double _M_processA_qqqq(double, double, const int *) const;
    double _M_processB_qqqq(double, double, const int *) const;
    double _M_processC_qqqq(double, double, const int *) const;
    
    /*     functions for selecting color configurations    */
    static double _S_select_color_gggg(shower_history<hhc>::iterator, double);
    static double _S_select_color_ggqQ(shower_history<hhc>::iterator, double);
    static double _S_select_color_gqgq(shower_history<hhc>::iterator, double);
    static double _S_select_color_gQgQ(shower_history<hhc>::iterator, double);
    static double _S_select_color_qggq(shower_history<hhc>::iterator, double);
    static double _S_select_color_QggQ(shower_history<hhc>::iterator, double);
    static double _S_select_color_qQgg(shower_history<hhc>::iterator, double);
    static double _S_select_color_Qqgg(shower_history<hhc>::iterator, double);
    static double _S_select_color_qQrR(shower_history<hhc>::iterator, double);
    static double _S_select_color_QqRr(shower_history<hhc>::iterator, double);
    static double _S_select_color_qrqr(shower_history<hhc>::iterator, double);
    static double _S_select_color_QRQR(shower_history<hhc>::iterator, double);
    static double _S_select_color_qRqR(shower_history<hhc>::iterator, double);
    static double _S_select_color_QrQr(shower_history<hhc>::iterator, double);
    
    static double _S_select_colorA_qQqQ(shower_history<hhc>::iterator, double);
    static double _S_select_colorB_qQqQ(shower_history<hhc>::iterator, double);
    static double _S_select_colorC_qQqQ(shower_history<hhc>::iterator, double);
    static double _S_select_colorA_QqQq(shower_history<hhc>::iterator, double);
    static double _S_select_colorB_QqQq(shower_history<hhc>::iterator, double);
    static double _S_select_colorC_QqQq(shower_history<hhc>::iterator, double);
    static double _S_select_colorA_qqqq(shower_history<hhc>::iterator, double);
    static double _S_select_colorB_qqqq(shower_history<hhc>::iterator, double);
    static double _S_select_colorC_qqqq(shower_history<hhc>::iterator, double);
    static double _S_select_colorA_QQQQ(shower_history<hhc>::iterator, double);
    static double _S_select_colorB_QQQQ(shower_history<hhc>::iterator, double);
    static double _S_select_colorC_QQQQ(shower_history<hhc>::iterator, double);
  };
  
}
#endif
